      subroutine cg(n, nmax, x,
     +     l, ldiag, jptrl,indrl,
     +     b, error, max_iter, sol, cg_iter, p, q, r, z,
     +     preconditioner, sqinvD)

c     Input variables

      integer n, nmax, max_iter, cg_iter
      integer jptrl(*),indrl(*), preconditioner
      double precision error, sqinvD(*)
      double precision l(*), ldiag(*), p(n),q(n),r(n),z(n)

c     Output variables

      double precision sol(n), x(nmax, *)
      double precision b(*)

c     Local variables

      integer i, j, k
      double precision beta,ptq, alpha, rho, rho_1, resid, normb
      double precision stime, ftime

c     Function declairation

      double precision ddot,dnrm2

      call dtimer(stime)

c     Initialize subroutine variables. 

      normb = dnrm2(n,b,1)
      do i = 1, n
         sol(i) = 0.0d0
      end do

       call dcopy(n,b,1,r,1)
       if (normb .eq. 0.0d0) then 
          normb = 1.0d0 
       endif
       call dcopy(n,r,1,p,1)
       resid = dnrm2(n,r,1)/normb 


       if (resid .le. error) then
          cg_iter = 0.0
          goto 100
       endif

       do 20 i = 1, max_iter


c         Diagonal or ICF preconditioning
c           z = L' \ (L \r)

          call dcopy(n,r,1,z,1)
          if (preconditioner .eq. 1) then
             call dstrsol(n,l,ldiag,jptrl, indrl,z,'N')
             call dstrsol(n,l,ldiag,jptrl, indrl,z,'T')
          endif
             
          rho = ddot(n, r, 1, z, 1) 
          if (i .eq. 1) then 
             call dcopy(n,z,1,p,1)
          else
             beta = rho / rho_1

c            p = z + beta * p

             call dscal(n,beta,p,1)
             call daxpy(n,1.0d0,z,1,p,1)
          endif

c         q = A*p

          do j = 1, n
             q(j) = 0.0d0
          end do
          do j = 1, n
             do k = 1, n
                q(j) = q(j) + sqinvD(k)*x(j,k)*sqinvD(j)*p(k)
             end do
          end do

c         Check for negative curvature.
          
          ptq = ddot(n,p,1,q,1)
          if (ptq .le. 0.0d0) then
             write (*,*) 'indefinite'
             cg_iter = dble(i)
             goto 100
          endif
          
          alpha = rho / ptq 
          call daxpy(n,alpha,p,1,sol,1)
          call daxpy(n,-alpha,q,1,r,1)
          resid = dnrm2(n,r,1)/normb 
          
          if (resid .le. error) then
             goto 100
          endif
          
          rho_1 = rho
 20      continue

 100     continue

         call dtimer(ftime)
         cg_iter = dble(i) 
         cgtime = ftime - stime
         write(*,*) 'CG iter = ', int(cg_iter), 'CG time = ', cgtime

         if (i .ge. max_iter) then 
            write(*,*) 'max iter of cg, error = ', resid
         endif
      end
