      program main

      character*50 fread, fwrite
      parameter (fread='in',fwrite='out')
      integer nnzmax, nmax
      parameter (nnzmax=1000000,nmax=2000)
      integer nread, nwrite, probfile
      parameter (nread=1,nwrite=2,probfile=8)
c     **********
c
c     Sample driver for the incomplete Cholesky decomposition.
c
c     Chih-Jen Lin, Nov 1998
c     Modified from ICFS
c     
c
c     **********
      character *60 prob_name

      integer i, j, nnzl, n
      integer info, iters1, iters2, maxiter
      integer p
      double precision ftime, stime
      integer lcol_ptr(nmax+1), lrow_ind(nnzmax)
      integer iwa(3*nmax)
      double precision alpha, rtol 
      double precision x(nmax, nmax), l(nnzmax),
     +     ldiag(nmax)
      double precision w(4*nmax), cgtime1, cgtime2, icftime,
     +      pretime1, pretime2, pretime3, sqinvD(nmax)
      double precision sol(nmax), b(nmax)

      external icf, cg, dtimer

      open (nread,file=fread,status='old')
      open (nwrite,file=fwrite)

      do while (1 .eq. 1)

         read (nread,*) prob_name
         if (prob_name .eq. 'STOP') stop
         read (nread, *) n
         open (probfile,file=prob_name)

         if (n .gt. nmax) then
            write (*,*) 'too big'
            stop
         endif

         write(*,*) 'read file', n
         do i = 1, n
            read(probfile, *) (x(j,i),j=1,n)
         end do

c        Find scaling D

         do i = 1, n
            sqinvD(i) = 1.0d0/sqrt(x(i,i))
         end do

         do i = 1, n
            b(i) = 0.0d0
            do j = 1, n
               b(i) = b(i) + x(i,j)
            end do
            b(i) = sqinvD(i) * b(i)  
         end do

         call dtimer(stime)
         alpha = 0.0d0
         p = int(10*(n**(1.0/3.0)))
         pretime1 = 0.0
         pretime2 = 0.0
         pretime3 = 0.0

         write(*,*) 'start icf'
 10      continue
         call icf(n,nmax, p,x, ldiag, l,lcol_ptr,lrow_ind,
     +        w, info, iwa, iwa(n+1), iwa(2*n+1),
     +        sqinvD, alpha, pretime1, pretime2, pretime3)
         if (info .lt. 0) then
            if (alpha .eq. 0.0d0) then
               alpha = 1.0d0
            else
               alpha = 2*alpha
            endif
            goto 10
         end if

         call dtimer(ftime)
         icftime = ftime - stime

         write(*,*) 'start cg'

         call dtimer(stime)

c        Solve Ax = b with the preconditioned 
c        conjugate gradient method.

         rtol = 1.0d-3
         maxiter = min(n,1000)
         call cg(n, nmax, x, l,ldiag,
     +        lcol_ptr,lrow_ind,b,rtol,maxiter,sol,
     +        iters1,w,w(n+1),w(2*n+1),
     +        w(3*n+1), 1, sqinvD)

         call dtimer(ftime)
         cgtime1 = ftime - stime

c        Diagonal preconditioner

         call dtimer(stime)
         call cg(n, nmax, x, l,ldiag,
     +        lcol_ptr,lrow_ind,b,rtol,maxiter,sol,
     +        iters2,w,w(n+1),w(2*n+1),
     +        w(3*n+1), 2, sqinvD)
         call dtimer(ftime)
         cgtime2 = ftime - stime

         nnzl = lcol_ptr(n+1) - 1 + n

         write (nwrite,1000) prob_name, n, nnzl, 
     +        iters1,alpha,pretime2/(pretime2+pretime3),
     +        pretime3/(pretime2+pretime3), icftime, cgtime1,
     +        iters2, cgtime2
          end do

      close(nread)
      close(nwrite)

      stop

 1000 format (1x,a23/,
     +        1x,'Order of the matrix                     ',3x,i10/,
     +        1x,'Number of nonzeroes in L                ',3x,i10/,
     +        1x,'Number of conjugate gradient iterations ',3x,i10/,
     +        1x,'Final shift                             ',3x,d10.3/,
     +        1x,'icf percentage                          ',3x,d10.3/,
     +        1x,'icf percentage                          ',3x,d10.3/,
     +        1x,'icf time                                ',3x,d10.3/,
     +        1x,'Time for pcg                            ',3x,d10.3/,
     +        1x,'Number of conjugate gradient iter(diag) ',3x,i10/,
     +        1x,'Time for pcg(diag)                      ',3x,d10.3/)

      end

